@@ -1,7 +1,6 @@
-import logging
 import os
 import shutil
-import urllib
+import urllib.error
 from dataclasses import dataclass
 from datetime import datetime
 from enum import IntEnum
@@ -11,10 +10,11 @@ from typing import Dict, List, Optional, Tuple
 
 import ramalama.go2jinja as go2jinja
 import ramalama.oci
-from ramalama.common import download_file, generate_sha256, verify_checksum
+from ramalama.arg_types import EngineArgs
+from ramalama.common import download_file, generate_sha256, perror, verify_checksum
+from ramalama.endian import EndianMismatchError, get_system_endianness
 from ramalama.gguf_parser import GGUFInfoParser, GGUFModelInfo
-
-LOGGER = logging.getLogger(__name__)
+from ramalama.logger import logger
 
 
 def sanitize_filename(filename: str) -> str:
@@ -25,6 +25,7 @@ class SnapshotFileType(IntEnum):
     Model = 1
     ChatTemplate = 2
     Other = 3
+    Mmproj = 4
 
 
 class SnapshotFile:
@@ -69,7 +70,14 @@ class LocalSnapshotFile(SnapshotFile):
         required: bool = True,
     ):
         super().__init__(
-            "", "", generate_sha256(content), name, type, should_show_progress, should_verify_checksum, required
+            "",
+            "",
+            generate_sha256(content),
+            name,
+            type,
+            should_show_progress,
+            should_verify_checksum,
+            required,
         )
         self.content = content
 
@@ -83,28 +91,35 @@ class LocalSnapshotFile(SnapshotFile):
 def validate_snapshot_files(snapshot_files: list[SnapshotFile]):
     model_files = []
     chat_template_files = []
+    mmproj_files = []
     for file in snapshot_files:
         if file.type == SnapshotFileType.Model:
             model_files.append(file)
         if file.type == SnapshotFileType.ChatTemplate:
             chat_template_files.append(file)
+        if file.type == SnapshotFileType.Mmproj:
+            mmproj_files.append(file)
 
     if len(model_files) > 1:
-        raise Exception(f"Only one model file supported, got {len(model_files)}: {model_files}")
+        raise ValueError(f"Only one model supported, got {len(model_files)}: {model_files}")
     if len(chat_template_files) > 1:
-        raise Exception(f"Only one chat template file supported, got {len(chat_template_files)}: {chat_template_files}")
+        raise ValueError(f"Only one chat template supported, got {len(chat_template_files)}: {chat_template_files}")
+    if len(mmproj_files) > 1:
+        raise ValueError(f"Only one mmproj supported, got {len(mmproj_files)}: {mmproj_files}")
 
 
 class RefFile:
     SEP = "---"
     MODEL_SUFFIX = "model"
     CHAT_TEMPLATE_SUFFIX = "chat"
+    MMPROJ_SUFFIX = "mmproj"
 
     def __init__(self):
         self.hash: str = ""
         self.filenames: list[str] = []
         self.model_name: str = ""
         self.chat_template_name: str = ""
+        self.mmproj_name: str = ""
         self._path: str = ""
 
     @property
@@ -129,6 +144,8 @@ class RefFile:
                     ref_file.model_name = parts[0]
                 if parts[1] == RefFile.CHAT_TEMPLATE_SUFFIX:
                     ref_file.chat_template_name = parts[0]
+                if parts[1] == RefFile.MMPROJ_SUFFIX:
+                    ref_file.mmproj_name = parts[0]
 
                 filename = file.readline().strip()
         return ref_file
@@ -141,6 +158,8 @@ class RefFile:
                 self.chat_template_name = ""
             if self.model_name == name:
                 self.model_name = ""
+            if self.mmproj_name == name:
+                self.mmproj_name = ""
 
     def serialize(self) -> str:
         lines = [self.hash]
@@ -150,6 +169,8 @@ class RefFile:
                 line = line + RefFile.MODEL_SUFFIX
             if filename == self.chat_template_name:
                 line = line + RefFile.CHAT_TEMPLATE_SUFFIX
+            if filename == self.mmproj_name:
+                line = line + RefFile.MMPROJ_SUFFIX
             lines.append(line)
         return "\n".join(lines)
 
@@ -172,14 +193,6 @@ DIRECTORY_NAME_REFS = "refs"
 DIRECTORY_NAME_SNAPSHOTS = "snapshots"
 
 
-class dotdict(dict):
-    """dot.notation access to dictionary attributes"""
-
-    __getattr__ = dict.get
-    __setattr__ = dict.__setitem__
-    __delattr__ = dict.__delitem__
-
-
 class GlobalModelStore:
     def __init__(
         self,
@@ -191,7 +204,7 @@ class GlobalModelStore:
     def path(self) -> str:
         return self._store_base_path
 
-    def list_models(self, engine: str, debug: bool, show_container: bool) -> Dict[str, List[ModelFile]]:
+    def list_models(self, engine: str, show_container: bool) -> Dict[str, List[ModelFile]]:
         models: Dict[str, List[ModelFile]] = {}
 
         for root, subdirs, _ in os.walk(self.path):
@@ -205,7 +218,8 @@ class GlobalModelStore:
                     model_source = parts[0]
                     model_path_without_source = f"{os.sep}".join(parts[1:])
 
-                    model_name = f"{model_source}://{model_path_without_source}:{ref_file_name}"
+                    separator = ":///" if model_source == "file" else "://"  # Use ':///' for file URLs, '://' otherwise
+                    model_name = f"{model_source}{separator}{model_path_without_source}:{ref_file_name}"
 
                     collected_files = []
                     for snapshot_file in ref_file.filenames:
@@ -217,11 +231,9 @@ class GlobalModelStore:
                             )
                             if not os.path.exists(blobs_partial_file_path):
                                 continue
-                            snapshot_file_path = blobs_partial_file_path
 
-                            # append indication for partial downloaded model
-                            if not model_name.endswith("(partial)"):
-                                is_partially_downloaded = True
+                            snapshot_file_path = blobs_partial_file_path
+                            is_partially_downloaded = True
 
                         last_modified = os.path.getmtime(snapshot_file_path)
                         file_size = os.path.getsize(snapshot_file_path)
@@ -231,14 +243,7 @@ class GlobalModelStore:
                     models[model_name] = collected_files
 
         if show_container:
-            oci_models = ramalama.oci.list_models(
-                dotdict(
-                    {
-                        "engine": engine,
-                        "debug": debug,
-                    }
-                )
-            )
+            oci_models = ramalama.oci.list_models(EngineArgs(engine=engine))
             for oci_model in oci_models:
                 name, modified, size = (oci_model["name"], oci_model["modified"], oci_model["size"])
                 # ramalama.oci.list_models provides modified as timestamp string, convert it to unix timestamp
@@ -407,6 +412,8 @@ class ModelStore:
                     ref_file.model_name = file.name
                 if file.type == SnapshotFileType.ChatTemplate:
                     ref_file.chat_template_name = file.name
+                if file.type == SnapshotFileType.Mmproj:
+                    ref_file.mmproj_name = file.name
             ref_file.write_to_file()
 
         snapshot_directory = self.get_snapshot_directory(snapshot_hash)
@@ -430,7 +437,7 @@ class ModelStore:
 
             if file.should_verify_checksum:
                 if not verify_checksum(dest_path):
-                    LOGGER.info(f"Checksum mismatch for blob {dest_path}, retrying download ...")
+                    logger.info(f"Checksum mismatch for blob {dest_path}, retrying download ...")
                     os.remove(dest_path)
                     file.download(dest_path, self.get_snapshot_directory(snapshot_hash))
                     if not verify_checksum(dest_path):
@@ -442,7 +449,7 @@ class ModelStore:
         ref_file.write_to_file()
 
     def _ensure_chat_template(self, model_tag: str, snapshot_hash: str, snapshot_files: list[SnapshotFile]):
-        model_file: SnapshotFile = None
+        model_file: SnapshotFile | None = None
         for file in snapshot_files:
             # Give preference to a chat template that has been specified in the file list
             if file.type == SnapshotFileType.ChatTemplate:
@@ -461,7 +468,7 @@ class ModelStore:
                     ]
                     self.update_snapshot(model_tag, snapshot_hash, files)
                 except Exception as ex:
-                    LOGGER.debug(f"Failed to convert Go Template to Jinja: {ex}")
+                    logger.debug(f"Failed to convert Go Template to Jinja: {ex}")
                 return
             if file.type == SnapshotFileType.Model:
                 model_file = file
@@ -496,15 +503,56 @@ class ModelStore:
                     LocalSnapshotFile(jinja_template, "chat_template_converted", SnapshotFileType.ChatTemplate)
                 )
             except Exception as ex:
-                LOGGER.debug(f"Failed to convert Go Template to Jinja: {ex}")
+                logger.debug(f"Failed to convert Go Template to Jinja: {ex}")
 
         self.update_snapshot(model_tag, snapshot_hash, files)
 
+    def _verify_endianness(self, model_tag: str):
+        ref_file = self.get_ref_file(model_tag)
+        if ref_file is None:
+            return
+
+        model_hash = self.get_blob_file_hash(ref_file.hash, ref_file.model_name)
+        model_path = self.get_blob_file_path(model_hash)
+
+        # only check endianness for gguf models
+        if not GGUFInfoParser.is_model_gguf(model_path):
+            return
+
+        model_endianness = GGUFInfoParser.get_model_endianness(model_path)
+        host_endianness = get_system_endianness()
+        if host_endianness != model_endianness:
+            raise EndianMismatchError(host_endianness, model_endianness)
+
+    def verify_snapshot(self, model_tag: str):
+        self._verify_endianness(model_tag)
+        self._store.verify_snapshot()
+
     def new_snapshot(self, model_tag: str, snapshot_hash: str, snapshot_files: list[SnapshotFile]):
         snapshot_hash = sanitize_filename(snapshot_hash)
-        self._prepare_new_snapshot(model_tag, snapshot_hash, snapshot_files)
-        self._download_snapshot_files(model_tag, snapshot_hash, snapshot_files)
-        self._ensure_chat_template(model_tag, snapshot_hash, snapshot_files)
+
+        try:
+            self._prepare_new_snapshot(model_tag, snapshot_hash, snapshot_files)
+            self._download_snapshot_files(model_tag, snapshot_hash, snapshot_files)
+            self._ensure_chat_template(model_tag, snapshot_hash, snapshot_files)
+        except urllib.error.HTTPError as ex:
+            perror(f"Failed to fetch required file: {ex}")
+            perror("Removing snapshot...")
+            self.remove_snapshot(model_tag)
+            raise ex
+        except Exception as ex:
+            perror(f"Failed to create new snapshot: {ex}")
+            perror("Removing snapshot...")
+            self.remove_snapshot(model_tag)
+            raise ex
+
+        try:
+            self.verify_snapshot(model_tag)
+        except EndianMismatchError as ex:
+            perror(f"Verification of snapshot failed: {ex}")
+            perror("Removing snapshot...")
+            self.remove_snapshot(model_tag)
+            raise ex
 
     def update_snapshot(self, model_tag: str, snapshot_hash: str, new_snapshot_files: list[SnapshotFile]) -> bool:
         validate_snapshot_files(new_snapshot_files)
@@ -524,6 +572,8 @@ class ModelStore:
                 ref_file.model_name = file.name
             if file.type == SnapshotFileType.ChatTemplate:
                 ref_file.chat_template_name = file.name
+            if file.type == SnapshotFileType.Mmproj:
+                ref_file.mmproj_name = file.name
 
         ref_file.write_to_file()
 
@@ -535,9 +585,9 @@ class ModelStore:
         try:
             if os.path.exists(blob_path) and Path(self.base_path) in blob_path.parents:
                 os.remove(blob_path)
-                LOGGER.debug(f"Removed blob for '{snapshot_file_path}'")
+                logger.debug(f"Removed blob for '{snapshot_file_path}'")
         except Exception as ex:
-            LOGGER.error(f"Failed to remove blob file '{blob_path}': {ex}")
+            logger.error(f"Failed to remove blob file '{blob_path}': {ex}")
 
     def remove_snapshot(self, model_tag: str):
         ref_file = self.get_ref_file(model_tag)
@@ -550,8 +600,11 @@ class ModelStore:
 
         # Remove snapshot directory
         snapshot_directory = self.get_snapshot_directory_from_tag(model_tag)
-        shutil.rmtree(snapshot_directory, ignore_errors=False)
+        shutil.rmtree(snapshot_directory, ignore_errors=True)
 
-        # Remove ref file
+        # Remove ref file, ignore if file is not found
         ref_file_path = self.get_ref_file_path(model_tag)
-        os.remove(ref_file_path)
+        try:
+            os.remove(ref_file_path)
+        except FileNotFoundError:
+            pass
